#==========================================================================================
# ce718 - Métodos Computacionalmente Intensivos                            Paulo Justiniano
# Metropolis-Hastings                                                          (08/08/2011)
#                                                                            Walmes Zeviani
#==========================================================================================

#------------------------------------------------------------------------------------------
# Independence Sampler - Amostrador independente
# *amostrador de Gibbs: retém todos os valores simulados das condicionais completas
# *amostrador de Gibbs: o resultado é uma cadeia (não independente) de valores
# *requer que a distribuição proponente contenha o suporte da distribuição alvo
# *realização da proponente não depende do valore assumido no passo anterior
# *a taxa de aceitação atual depende do último valor gerado

#------------------------------------------------------------------------------------------
# ilustrando o procedimento (com um exemplo bem ridículo)
# obter realizações de uma distribuição Gamma(alpha, beta)
# definir a distribuição proponente q_X(x,.)

q_X <- function(alpha, beta){
  ## alpha e beta: escalares que são parâmetros da distribuição alvo
  mu <- alpha/beta            # esperança da distribuição alvo
  sig <- sqrt(alpha/(beta^2)) # variância da distribuiçaõ alvo
  rnorm(1,mu,sig)      # distribuição proponente é a normal
  ## retorna uma realização da distribuição proponente
}

iidsampler <- function(nsim, alpha, beta, xini=alpha/beta){
  ## nsim: escalar número de simulações
  ## alpha e beta: escalares esperança de variância da distribuição alvo
  out <- rep(NA, nsim)
  mu <- alpha/beta            # esperança da distribuição alvo
  sig <- sqrt(alpha/(beta^2)) # variância da distribuiçaõ alvo
  out[1] <- xini
  for(i in 2:nsim){
    can <- q_X(alpha, beta) # realização da distribuição proponente
    ratio <- (dgamma(can, alpha, beta)/dgamma(out[i-1], alpha, beta))/
      (dnorm(can, mu, sig)/dnorm(out[i-1], mu, sig)) # cálculo da razão de aceitação
    u <- runif(1)
    if(u<ratio) out[i] <- can else out[i] <- out[i-1]
    ## incluir contador da aceitação
  }
  return(out)
}

n <- 5000; alpha <- 2; beta <- 1 # trocar valores de alpha
x <- iidsampler(n, alpha, beta, xini=2) # trocar valor de xini=9
plot(x, type="l")
acf(x)

hist(x, freq=FALSE)
lines(density(x), col=2)
curve(dgamma(x, alpha, beta), add=TRUE, col=4)
hist(x[-c(1:4000)], freq=FALSE, add=TRUE)

ks.test(x, "pgamma", shape=alpha, rate=beta)

#------------------------------------------------------------------------------------------
# Ramdon Walk - Caminhada Aleatória
# *explora a vizinha do valor atual para propor novos valores;
# *é um caso particular do algoritmo Metropolis onde a proponente deve ser simétrica
# *no RW a proponente além de simétrica é função da diferença entre y e x[t], g(y-x[t])
# *com isso, o ratio só depende de avaliarmos a razão das avaliações da dist alvo, f(.)

#------------------------------------------------------------------------------------------
# ilustrando o procedimento (com um exemplo bem ridículo)
# obter realizações de uma distribuição Normal(0, 1)
# definir a distribuição proponente q_X(x,.) como sendo a uniforme(-delta,delta)

q_X <- function(delta, xi){
  ## delta e xi: escalares parâmetros da distribuição proponente
  runif(1, xi-delta, xi+delta)      # distribuição proponente é a uniforme
  ## retorna uma realização da distribuição proponente
}

rwsampler <- function(nsim, xiini, delta, mu, sigma){
  out <- rep(NA, nsim)
  out[1] <- xiini
  for(i in 2:nsim){
    can <- q_X(delta, xi=out[i-1])
    ratio <- dnorm(can, mu, sigma)/dnorm(out[i-1], mu, sigma)
    u <- runif(1)
    if(u<ratio) out[i] <- can else out[i] <- out[i-1]
  }
  return(out)
}

# usar como alvo a gamma e tratar os candidatos=0

mu <- 0; sigma <- 1
x <- rwsampler(nsim=5000, xiini=0, delta=5, mu, sigma) # trocar valor de delta
plot(x, type="l")
acf(x)

tinn <- seq(1,5000,100)
plot(x[tinn], type="l")
acf(x[tinn])

hist(x, freq=FALSE)
hist(x[tinn], freq=FALSE, add=TRUE)
lines(density(x), col=2)
curve(dnorm(x, mu, sigma), add=TRUE, col=4)

ks.test(x, "pnorm", mean=mu, sd=sigma)

#------------------------------------------------------------------------------------------
# Randon Walk
# obter uma amostra aleatória da função pi(x) =~ exp(-x^2)+0.5*exp(-(x-5)^2), x in R

curve(exp(-x^2)+0.5*exp(-(x-5)^2), -10, 20) # distribuição bimodal, não integra 1
norm <- 1/integrate(function(x){exp(-x^2)+0.5*exp(-(x-5)^2)}, -Inf, Inf)$value

integrate(function(x){norm*(exp(-x^2)+0.5*exp(-(x-5)^2))}, -Inf, Inf)$value

rwsampler <- function(nsim, xini, sigma){
  out <- rep(NA, nsim)
  out[1] <- xini
  for(i in 2:nsim){
    can <- rnorm(1, out[i-1], sigma) # proponente é a normal
    ratio <- (exp(-can^2)+0.5*exp(-(can-5)^2))/
      (exp(-out[i-1]^2)+0.5*exp(-(out[i-1]-5)^2))
    ## PJ taste
    #ratio <- exp(log((exp(-can^2)+0.5*exp(-(can-5)^2)))-
    #             log((exp(-out[i-1]^2)+0.5*exp(-(out[i-1]-5)^2))))
    u <- runif(1)
    if(u<ratio) out[i] <- can else out[i] <- out[i-1]
  }
  return(out)
}

x <- rwsampler(nsim=5000, xini=100, sigma=3) # trocar valor de delta
plot(x, type="l")
acf(x)

hist(x, freq=FALSE)
rug(x)
lines(density(x), col=2)
curve(norm*(exp(-x^2)+0.5*exp(-(x-5)^2)), -10, 20, add=TRUE)

#------------------------------------------------------------------------------------------
# Metropolis-Hastings
# *proponente não precisa ser simétrica
# *no ratio entra a correção de Hastings para a não simetria da proponente
# *o resto é igual

#------------------------------------------------------------------------------------------
# aplicação de inferência Bayesiana
# yi ~Poisson(lambda), verossimilhança Poisson
# priori lambda ~Normal(mu, sigma²)

set.seed(1234)
n <- 30; lambda <- 5
y <- rpois(n, lambda)  # gerando as 10 realizações
mu <- 5; sigma2 <- 3   # hiperparametros da priori

# proponentes
lambda <- 5; k <- 1
curve(dnorm(x, lambda, sqrt(k)), 0, 10)
curve(dgamma(x, lambda, k), 0, 10, add=TRUE)

post <- function(nsim, y, lambdaini, mu, sigma2, k, prop="normal"){
  out <-rep(NA, nsim)
  out[1] <- lambdaini
  if(prop=="normal"){
    q_X <- function(mu, k){ rnorm(1, mu, sqrt(k))}
  }
  if(prop=="gamma"){
    q_X <- function(mu, k){ rgamma(1, mu, k)}
  }
  for(i in 2:nsim){
    can <- q_X(out[i-1], k)
    ##can <- rnorm(1, out[i-1], k)
    num <- sum(dpois(y, lambda=can, log=TRUE))+dnorm(can, mu, sqrt(sigma2), log=TRUE)
    den <- sum(dpois(y, lambda=out[i-1], log=TRUE))+dnorm(out[i-1], mu, sqrt(sigma2), log=TRUE)
    ratio <- exp(num)/exp(den) # trocar para exp(num-den)
    u <- runif(1)
    if(u<ratio) out[i] <- can else out[i] <- out[i-1]
  }
  return(out)
}

x <- post(nsim=10000, y, lambdaini=3, mu=4, sigma2=1, k=1, prop="normal")
#x <- post(nsim=10000, y, lambdaini=3, mu=4, sigma2=1, k=1, prop="gamma")
plot(x, type="l")
acf(x)

hist(x, freq=FALSE)
rug(x)
lines(density(x), col=2)
abline(v=c(mean(x), quantile(x, p=c(25,500,975)/1000)))

summary(x)
summary(y)

#------------------------------------------------------------------------------------------
